@@ -54,11 +54,14 @@ struct _GsmShellPrivate
         DBusGProxy      *bus_proxy;
         DBusGProxy      *proxy;
         DBusGProxy      *end_session_dialog_proxy;
+        GsmStore        *inhibitors;
 
         guint32          is_running : 1;
 
         DBusGProxyCall  *end_session_open_call;
         GsmShellEndSessionDialogType end_session_dialog_type;
+
+        guint            update_idle_id;
 };
 
 enum {
@@ -128,7 +131,7 @@ gsm_shell_class_init (GsmShellClass *shell_class)
 
         param_spec = g_param_spec_boolean ("is-running",
                                            "Is running",
-                                           "Whether the session is running to ConsoleKit",
+                                           "Whether GNOME Shell is running in the session",
                                            FALSE,
                                            G_PARAM_READABLE);
 
@@ -267,6 +270,8 @@ gsm_shell_ensure_connection (GsmShell  *shell,
                 }
         }
 
+        g_debug ("GsmShell: Connected to the shell");
+
         is_running = TRUE;
 
  out:
@@ -276,6 +281,8 @@ gsm_shell_ensure_connection (GsmShell  *shell,
         }
 
         if (!is_running) {
+                g_debug ("GsmShell: Not connected to the shell");
+
                 if (shell->priv->bus_connection == NULL) {
                         if (shell->priv->bus_proxy != NULL) {
                                 g_object_unref (shell->priv->bus_proxy);
@@ -319,17 +326,9 @@ gsm_shell_on_name_owner_changed (DBusGProxy    *bus_proxy,
 static void
 gsm_shell_init (GsmShell *shell)
 {
-        GError *error;
-
         shell->priv = GSM_SHELL_GET_PRIVATE (shell);
 
-        error = NULL;
-
-        if (!gsm_shell_ensure_connection (shell, &error)) {
-                g_debug ("Could not connect to the shell: %s",
-                         error->message);
-                g_error_free (error);
-        }
+        gsm_shell_ensure_connection (shell, NULL);
 }
 
 static void
@@ -367,6 +366,8 @@ gsm_shell_finalize (GObject *object)
 
         parent_class = G_OBJECT_CLASS (gsm_shell_parent_class);
 
+        g_object_unref (shell->priv->inhibitors);
+
         gsm_shell_disconnect_from_bus (shell);
 
         if (parent_class->finalize != NULL) {
@@ -446,6 +447,11 @@ on_open_finished (DBusGProxy     *proxy,
                                      G_TYPE_INVALID);
         shell->priv->end_session_open_call = NULL;
 
+        if (shell->priv->update_idle_id != 0) {
+                g_source_remove (shell->priv->update_idle_id);
+                shell->priv->update_idle_id = 0;
+        }
+
         if (!res) {
                 g_warning ("Unable to open shell end session dialog");
                 g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_OPEN_FAILED], 0);
@@ -459,6 +465,11 @@ static void
 on_end_session_dialog_canceled (DBusGProxy *proxy,
                                 GsmShell   *shell)
 {
+        if (shell->priv->update_idle_id != 0) {
+                g_source_remove (shell->priv->update_idle_id);
+                shell->priv->update_idle_id = 0;
+        }
+
         g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CANCELED], 0);
 }
 
@@ -466,6 +477,11 @@ static void
 on_end_session_dialog_confirmed (DBusGProxy *proxy,
                                  GsmShell   *shell)
 {
+        if (shell->priv->update_idle_id != 0) {
+                g_source_remove (shell->priv->update_idle_id);
+                shell->priv->update_idle_id = 0;
+        }
+
         g_signal_emit (G_OBJECT (shell), signals[END_SESSION_DIALOG_CONFIRMED], 0);
 }
 
@@ -479,11 +495,29 @@ on_end_session_dialog_proxy_destroyed (DBusGProxy *proxy,
         }
 }
 
+static gboolean
+on_need_end_session_dialog_update (GsmShell *shell)
+{
+        gsm_shell_open_end_session_dialog (shell,
+                                           shell->priv->end_session_dialog_type,
+                                           shell->priv->inhibitors);
+        return FALSE;
+}
+
+static void
+queue_end_session_dialog_update (GsmShell *shell)
+{
+        if (shell->priv->update_idle_id != 0)
+                return;
+
+        shell->priv->update_idle_id = g_idle_add ((GSourceFunc) on_need_end_session_dialog_update,
+                                                  shell);
+}
+
 gboolean
 gsm_shell_open_end_session_dialog (GsmShell *shell,
                                    GsmShellEndSessionDialogType type,
-                                   GsmStore *inhibitors,
-                                   GsmStore *clients)
+                                   GsmStore *inhibitors)
 {
         DBusGProxyCall  *call;
         GPtrArray *inhibitor_array;
@@ -562,6 +596,15 @@ gsm_shell_open_end_session_dialog (GsmShell *shell,
                 return FALSE;
         }
 
+        shell->priv->inhibitors = g_object_ref (inhibitors);
+        g_signal_connect_swapped (inhibitors, "added",
+                                  G_CALLBACK (queue_end_session_dialog_update),
+                                  shell);
+
+        g_signal_connect_swapped (inhibitors, "removed",
+                                  G_CALLBACK (queue_end_session_dialog_update),
+                                  shell);
+
         shell->priv->end_session_open_call = call;
         shell->priv->end_session_dialog_type = type;
 
